@@ -230,6 +230,7 @@ import warnings
 from oslo_config import cfg
 from oslo_context import context
 from oslo_serialization import jsonutils
+from oslo_utils import strutils
 import six
 import yaml
 
@@ -295,6 +296,15 @@ RuleCheck = _checks.RuleCheck
 """Recursively checks credentials based on the defined rules."""
 
 
+WARN_JSON = ("JSON formatted policy_file support is deprecated since "
+             "Victoria release. You need to use YAML format which "
+             "will be default in future. You can use "
+             "``oslopolicy-convert-json-to-yaml`` tool to convert existing "
+             "JSON-formatted policy file to YAML-formatted in backward "
+             "compatible way: https://docs.openstack.org/oslo.policy/"
+             "latest/cli/oslopolicy-convert-json-to-yaml.html.")
+
+
 class PolicyNotAuthorized(Exception):
     """Default exception raised for policy enforcement failure."""
 
@@ -368,6 +378,9 @@ def parse_file_contents(data):
         # by jsonutils.loads() first. In case of failure yaml.safe_load()
         # will be used instead.
         parsed = jsonutils.loads(data)
+        # NOTE(gmann): If policy file is loaded in JSON format means
+        # policy_file is JSON formatted so log warning.
+        LOG.warning(WARN_JSON)
     except ValueError:
         try:
             parsed = yaml.safe_load(data)
@@ -502,6 +515,14 @@ class Enforcer(object):
         self._policy_dir_mtimes = {}
         self._file_cache = {}
         self._informed_no_policy_file = False
+        # NOTE(gmann): This flag will suppress the warning for
+        # policies changing their default check_str that have
+        # not been overridden by operators. This does not affect the
+        # warning for policy changed their name or deprecated
+        # for removal.
+        self.suppress_default_change_warnings = False
+        # FOR TESTING ONLY
+        self.suppress_deprecation_warnings = False
 
     def set_rules(self, rules, overwrite=True, use_conf=False):
         """Create a new :class:`Rules` based on the provided dict of rules.
@@ -537,6 +558,8 @@ class Enforcer(object):
         self.registered_rules = {}
         self.file_rules = {}
         self._informed_no_policy_file = False
+        self.suppress_default_change_warnings = False
+        self.suppress_deprecation_warnings = False
 
     def load_rules(self, force_reload=False):
         """Loads policy_path's rules.
@@ -549,6 +572,8 @@ class Enforcer(object):
         if force_reload:
             self.use_conf = force_reload
 
+        policy_file_rules_changed = False
+
         if self.use_conf:
             if not self.policy_path:
                 try:
@@ -560,80 +585,37 @@ class Enforcer(object):
                         self._informed_no_policy_file = True
 
             if self.policy_path:
-                self._load_policy_file(self.policy_path, force_reload,
-                                       overwrite=self.overwrite)
+                # If the policy file rules have changed any policy.d rules
+                # also need to be reapplied on top of that change.
+                policy_file_rules_changed = self._load_policy_file(
+                    self.policy_path,
+                    force_reload,
+                    overwrite=self.overwrite
+                )
+
+            force_reload_policy_dir = force_reload
+            if policy_file_rules_changed:
+                force_reload_policy_dir = True
+
             for path in self.conf.oslo_policy.policy_dirs:
                 try:
                     path = self._get_policy_path(path)
                 except cfg.ConfigFilesNotFoundError:
                     continue
-                if (force_reload or self._is_directory_updated(
-                        self._policy_dir_mtimes, path)):
-                    self._walk_through_policy_directory(path,
-                                                        self._load_policy_file,
-                                                        force_reload, False)
+                if (self._is_directory_updated(self._policy_dir_mtimes, path)
+                        or force_reload_policy_dir):
+                    self._walk_through_policy_directory(
+                        path,
+                        self._load_policy_file,
+                        force_reload_policy_dir, False
+                    )
 
             for default in self.registered_rules.values():
-                if default.deprecated_rule:
-                    deprecated_msg = (
-                        'Policy "%(old_name)s":"%(old_check_str)s" was '
-                        'deprecated in %(release)s in favor of "%(name)s":'
-                        '"%(check_str)s". Reason: %(reason)s. Either ensure '
-                        'your deployment is ready for the new default or '
-                        'copy/paste the deprecated policy into your policy '
-                        'file and maintain it manually.' % {
-                            'old_name': default.deprecated_rule.name,
-                            'old_check_str': default.deprecated_rule.check_str,
-                            'release': default.deprecated_since,
-                            'name': default.name,
-                            'check_str': default.check_str,
-                            'reason': default.deprecated_reason
-                        }
-                    )
-                    if default.deprecated_rule.name != default.name and (
-                            default.deprecated_rule.name in self.rules):
-                        # Print a warning because the actual policy name is
-                        # changing. If deployers are relying on an override for
-                        # foo:bar and it's getting renamed to foo:create_bar
-                        # then they need to be able to see that before they
-                        # roll out the next release.
-                        warnings.warn(deprecated_msg)
-                    if (default.deprecated_rule.check_str !=
-                            default.check_str and default.name not in
-                            self.rules):
-                        # In this case, the default check_str is changing. We
-                        # need to let operators know that this is going to
-                        # change. If they don't want to override it, they are
-                        # going to have to make sure the right infrastructure
-                        # exists before they upgrade. This overrides the new
-                        # check with an OrCheck that combines the new and old
-                        # check_str attributes from the new and deprecated
-                        # policies. This will make it so that deployments don't
-                        # break on upgrade, but they receive log messages
-                        # telling them stuff is going to change if they don't
-                        # maintain the policy manually or add infrastructure to
-                        # their deployment to support the new policy.
-                        default.check = _parser.parse_rule(
-                            default.check_str + ' or ' +
-                            default.deprecated_rule.check_str
-                        )
-                        warnings.warn(deprecated_msg)
-                if default.deprecated_for_removal and (
-                        default.name in self.file_rules):
-                    # If a policy is going to be removed altogether, then we
-                    # need to make sure we let operators know so they can clean
-                    # up their policy files, if they are overriding it.
-                    warnings.warn(
-                        'Policy "%(policy)s":"%(check_str)s" was '
-                        'deprecated for removal in %(release)s. Reason: '
-                        '%(reason)s. Its value may be silently ignored in '
-                        'the future.' % {
-                            'policy': default.name,
-                            'check_str': default.check_str,
-                            'release': default.deprecated_since,
-                            'reason': default.deprecated_reason
-                        }
-                    )
+                if default.deprecated_for_removal:
+                    self._emit_deprecated_for_removal_warning(default)
+                elif default.deprecated_rule:
+                    self._handle_deprecated_rule(default)
+
                 if default.name not in self.rules:
                     self.rules[default.name] = default.check
 
@@ -667,6 +649,99 @@ class Enforcer(object):
 
         return not violation
 
+    def _emit_deprecated_for_removal_warning(self, default):
+        # If the policy is being removed completely, we need to let operators
+        # know that the policy is going to be silently ignored in the future
+        # and they can remove it from their overrides since it isn't being
+        # replaced by another policy.
+        if not self.suppress_deprecation_warnings and \
+                default.name in self.file_rules:
+            warnings.warn(
+                'Policy "%(policy)s":"%(check_str)s" was deprecated for '
+                'removal in %(release)s. Reason: %(reason)s. Its value may be '
+                'silently ignored in the future.' % {
+                    'policy': default.name,
+                    'check_str': default.check_str,
+                    'release': default.deprecated_since,
+                    'reason': default.deprecated_reason
+                }
+            )
+
+    def _handle_deprecated_rule(self, default):
+        """Handle cases where a policy rule has been deprecated.
+
+        :param default: an instance of RuleDefault that contains an instance of
+                        DeprecatedRule
+        """
+
+        deprecated_rule = default.deprecated_rule
+
+        deprecated_msg = (
+            'Policy "%(old_name)s":"%(old_check_str)s" was deprecated in '
+            '%(release)s in favor of "%(name)s":"%(check_str)s". Reason: '
+            '%(reason)s. Either ensure your deployment is ready for the new '
+            'default or copy/paste the deprecated policy into your policy '
+            'file and maintain it manually.' % {
+                'old_name': deprecated_rule.name,
+                'old_check_str': deprecated_rule.check_str,
+                'release': default.deprecated_since,
+                'name': default.name,
+                'check_str': default.check_str,
+                'reason': default.deprecated_reason
+            }
+        )
+
+        # Print a warning because the actual policy name is changing. If
+        # operators are relying on an override for foo:bar and it's getting
+        # renamed to foo:create_bar then they need to be able to see that
+        # before they roll out the next release. If the policy name is in
+        # self.file_rules, we know that it's being overridden.
+        if deprecated_rule.name != default.name and (
+                deprecated_rule.name in self.file_rules):
+
+            if not self.suppress_deprecation_warnings:
+                warnings.warn(deprecated_msg)
+
+            # If the deprecated policy is being overridden and doesn't match
+            # the default deprecated policy, override the new policy's default
+            # with the old check string. This should prevents unwanted exposure
+            # to APIs on upgrade.
+            # There's one exception to this: When we generate a sample policy,
+            # we set the deprecated rule name to reference the new rule. If we
+            # see that the deprecated override rule is just the new rule, then
+            # we shouldn't mess with it.
+            if (self.file_rules[deprecated_rule.name].check
+                    != _parser.parse_rule(deprecated_rule.check_str) and
+                    str(self.file_rules[deprecated_rule.name].check)
+                    != 'rule:%s' % default.name):
+                if default.name not in self.file_rules.keys():
+                    self.rules[default.name] = self.file_rules[
+                        deprecated_rule.name
+                    ].check
+
+        # In this case, the default check string is changing. We need to let
+        # operators know that this is going to change. If they don't want to
+        # override it, they are going to have to make sure the right
+        # infrastructure exists before they upgrade. This overrides the new
+        # check with an OrCheck that combines the new and old check_str
+        # attributes from the new and deprecated policies. This will make it so
+        # that deployments don't break on upgrade, but they receive log
+        # messages telling them stuff is going to change if they don't maintain
+        # the policy manually or add infrastructure to their deployment to
+        # support the new policy.
+        # If the enforce_new_defaults flag is True, do not add OrCheck to the
+        # old check_str and enforce only the new defaults.
+        if (not self.conf.oslo_policy.enforce_new_defaults
+                and deprecated_rule.check_str != default.check_str
+                and default.name not in self.file_rules):
+
+            default.check = OrCheck([_parser.parse_rule(cs) for cs in
+                                     [default.check_str,
+                                      deprecated_rule.check_str]])
+            if not (self.suppress_deprecation_warnings
+                    or self.suppress_default_change_warnings):
+                warnings.warn(deprecated_msg)
+
     def _undefined_check(self, check):
         '''Check if a RuleCheck references an undefined rule.'''
         if isinstance(check, RuleCheck):
@@ -722,6 +797,8 @@ class Enforcer(object):
         # is in the cache
         mtime = 0
         if os.path.exists(path):
+            if not os.path.isdir(path):
+                raise ValueError('{} is not a directory'.format(path))
             # Make a list of all the files
             files = [path] + [os.path.join(path, file) for file in
                               os.listdir(path)]
@@ -756,18 +833,44 @@ class Enforcer(object):
         if overwrite:
             self.file_rules = {}
         parsed_file = parse_file_contents(data)
+        redundant_file_rules = []
         for name, check_str in parsed_file.items():
-            self.file_rules[name] = RuleDefault(name, check_str)
+            file_rule = RuleDefault(name, check_str)
+            self.file_rules[name] = file_rule
+            reg_rule = self.registered_rules.get(name)
+            if (reg_rule and (file_rule == reg_rule)):
+                redundant_file_rules.append(name)
+        if redundant_file_rules:
+            # NOTE(gmann): Log warning for redundant file rules which
+            # can be detected via 'oslopolicy-list-redundant' tool too.
+            LOG.warning("Policy Rules %(names)s specified in policy files "
+                        "are the same as the defaults provided by the "
+                        "service. You can remove these rules from policy "
+                        "files which will make maintenance easier. You can "
+                        "detect these redundant rules by "
+                        "``oslopolicy-list-redundant`` tool also.",
+                        {'names': redundant_file_rules})
 
     def _load_policy_file(self, path, force_reload, overwrite=True):
+        """Load policy rules from the specified policy file.
+
+        :param str path: A path of the policy file to load rules from.
+        :param bool force_reload: Forcefully reload the policy file content.
+        :param bool overwrite: Replace policy rules instead of updating them.
+        :return: A bool indicating whether rules have been changed or not.
+        :rtype: bool
+        """
+        rules_changed = False
         reloaded, data = _cache_handler.read_cached_file(
             self._file_cache, path, force_reload=force_reload)
         if reloaded or not self.rules:
             rules = Rules.load(data, self.default_rule)
             self.set_rules(rules, overwrite=overwrite, use_conf=True)
+            rules_changed = True
             self._record_file_rules(data, overwrite)
             self._loaded_files.append(path)
             LOG.debug('Reloaded policy file: %(path)s', {'path': path})
+        return rules_changed
 
     def _get_policy_path(self, path):
         """Locate the policy YAML/JSON data file/path.
@@ -799,8 +902,7 @@ class Enforcer(object):
                             operated on as possible. The target
                             argument should be a dict instance or an
                             instance of a class that fully supports
-                            the Mapping abstract base class and deep
-                            copying.
+                            the Mapping abstract base class.
         :param dict creds: As much information about the user performing the
                            action as possible. This parameter can also be an
                            instance of ``oslo_context.context.RequestContext``.
@@ -838,6 +940,27 @@ class Enforcer(object):
             )
             raise InvalidContextObject(msg)
 
+        if LOG.isEnabledFor(logging.DEBUG):
+            try:
+                creds_dict = strutils.mask_dict_password(creds)
+                creds_msg = jsonutils.dumps(creds_dict,
+                                            skipkeys=True, sort_keys=True)
+            except Exception as e:
+                creds_msg = ('cannot format data, exception: %(exp)s' %
+                             {'exp': e})
+
+            try:
+                target_dict = strutils.mask_dict_password(target)
+                target_msg = jsonutils.dumps(target_dict,
+                                             skipkeys=True, sort_keys=True)
+            except Exception as e:
+                target_msg = ('cannot format data, exception: %(exp)s' %
+                              {'exp': e})
+
+            LOG.debug('enforce: rule=%s creds=%s target=%s',
+                      rule.__class__ if isinstance(rule, _checks.BaseCheck)
+                      else '"%s"' % rule, creds_msg, target_msg)
+
         # Allow the rule to be a Check tree
         if isinstance(rule, _checks.BaseCheck):
             # If the thing we're given is a Check, we don't know the
@@ -860,17 +983,20 @@ class Enforcer(object):
                 # If the rule doesn't exist, fail closed
                 result = False
             else:
+                # NOTE(moguimar): suppressing [B105:hardcoded_password_string]
+                #                 as token_scope is not actually a hardcoded
+                #                 token.
+
                 # Check the scope of the operation against the possible scope
                 # attributes provided in `creds`.
                 if creds.get('system'):
-                    token_scope = 'system'
+                    token_scope = 'system'  # nosec
+                elif creds.get('domain_id'):
+                    token_scope = 'domain'  # nosec
                 else:
-                    # If the token isn't system-scoped then we're dealing with
-                    # either a domain-scoped token or a project-scoped token.
-                    # From a policy perspective, both are "project" operations.
-                    # Whether or not the project is a domain depends on where
-                    # it sits in the hierarchy.
-                    token_scope = 'project'
+                    # If the token isn't system-scoped or domain-scoped then
+                    # we're dealing with a project-scoped token.
+                    token_scope = 'project'  # nosec
 
                 registered_rule = self.registered_rules.get(rule)
                 if registered_rule and registered_rule.scope_types:
